@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2014-2017, Nils Christopher Brause
+ * Copyright (c) 2014-2019, Nils Christopher Brause, Philipp Kerling, Zsolt Bölöny
  * All rights reserved.
  * 
  * Redistribution and use in source and binary forms, with or without
@@ -203,12 +203,6 @@ public:
         has_keyboard = capability & seat_capability::keyboard;
         has_pointer = capability & seat_capability::pointer;
       };
-    display.roundtrip();
-
-    if(!has_keyboard)
-      throw std::runtime_error("No keyboard found.");
-    if(!has_pointer)
-      throw std::runtime_error("No pointer found.");
 
     // create a surface
     surface = compositor.create_surface();
@@ -218,6 +212,7 @@ public:
       {
         xdg_wm_base.on_ping() = [&] (uint32_t serial) { xdg_wm_base.pong(serial); };
         xdg_surface = xdg_wm_base.get_xdg_surface(surface);
+        xdg_surface.on_configure() = [&] (uint32_t serial) { xdg_surface.ack_configure(serial); };
         xdg_toplevel = xdg_surface.get_toplevel();
         xdg_toplevel.set_title("Window");
         xdg_toplevel.on_close() = [&] () { running = false; };
@@ -229,8 +224,16 @@ public:
         shell_surface.set_title("Window");
         shell_surface.set_toplevel();
       }
+    surface.commit();
+
+    display.roundtrip();
 
     // Get input devices
+    if(!has_keyboard)
+      throw std::runtime_error("No keyboard found.");
+    if(!has_pointer)
+      throw std::runtime_error("No pointer found.");
+
     pointer = seat.get_pointer();
     keyboard = seat.get_keyboard();
 
