@@ -1,5 +1,5 @@
 /*
- *  Copyright (c) 2014-2017 Nils Christopher Brause
+ *  Copyright (c) 2014-2019 Nils Christopher Brause, Philipp Kerling, Bernd Kuhls
  *
  *  This program is free software: you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -20,8 +20,10 @@
 #include <list>
 #include <set>
 #include <sstream>
+#include <vector>
 #include <cctype>
 #include <cmath>
+#include <stdexcept>
 
 #include "pugixml.hpp"
 
@@ -31,14 +33,22 @@ std::list<std::string> interface_names;
 
 struct element_t
 {
+  std::string name;
   std::string summary;
   std::string description;
+  static const std::set<std::string> keywords;
+
+  std::string sanitise(std::string str)
+  {
+    if(keywords.count(str))
+      return "_" + str;
+    return str;
+  }
 };
 
 struct argument_t : public element_t
 {
   std::string type;
-  std::string name;
   std::string interface;
   std::string enum_iface;
   std::string enum_name;
@@ -105,13 +115,12 @@ struct argument_t : public element_t
 
   std::string print_argument()
   {
-    return print_type() + " " + name;
+    return print_type() + " " + sanitise(name);
   }
 };
 
 struct event_t : public element_t
 {
-  std::string name;
   std::list<argument_t> args;
   int since;
 
@@ -124,7 +133,7 @@ struct event_t : public element_t
     if(args.size())
       ss.str(ss.str().substr(0, ss.str().size()-2));
     ss.seekp(0, std::ios_base::end);
-    ss << ")> " << name << ";";
+    ss << ")> " << sanitise(name) << ";";
     return ss.str();
   }
 
@@ -132,11 +141,11 @@ struct event_t : public element_t
   {
     std::stringstream ss;
     ss << "    case " << opcode << ":" << std::endl
-       << "      if(events->" << name << ") events->" << name << "(";
+       << "      if(events->" << sanitise(name) << ") events->" << sanitise(name) << "(";
 
     int c = 0;
     for(auto &arg : args)
-      if(arg.enum_name != "")
+      if(arg.enum_name != "" && arg.type != "array")
         ss << arg.print_type() << "(args[" << c++ << "].get<" << arg.print_enum_wire_type() << ">()), ";
       else if(arg.interface != "")
         ss << arg.print_type() << "(args[" << c++ << "].get<proxy_t>()), ";
@@ -180,7 +189,7 @@ struct event_t : public element_t
     ss.seekp(0, std::ios_base::end);
     ss << ")> &" + interface_name + "_t::on_" + name + "()" << std::endl
        << "{" << std::endl
-       << "  return std::static_pointer_cast<events_t>(get_events())->" + name + ";" << std::endl
+       << "  return std::static_pointer_cast<events_t>(get_events())->" + sanitise(name) + ";" << std::endl
        << "}" << std::endl;
     return ss.str();
   }
@@ -219,7 +228,7 @@ struct request_t : public event_t
                  << "      \\param version Interface version" << std::endl;
           }
         else
-          ss << "      \\param " << arg.name << " " << arg.summary << std::endl;
+          ss << "      \\param " << sanitise(arg.name) << " " << arg.summary << std::endl;
       }
     ss << description << std::endl
        << "  */" << std::endl;
@@ -228,7 +237,7 @@ struct request_t : public event_t
       ss << "  void ";
     else
       ss << "  " << ret.print_type() << " ";
-    ss << name << "(";
+    ss << sanitise(name) << "(";
 
     for(auto &arg : args)
       if(arg.type == "new_id")
@@ -245,7 +254,7 @@ struct request_t : public event_t
     ss << ");" << std::endl;
 
     ss << std::endl
-       << "  /** \\brief Minimum protocol version required for the \\ref " << name << " function" << std::endl
+       << "  /** \\brief Minimum protocol version required for the \\ref " << sanitise(name) << " function" << std::endl
        << "  */" << std::endl
        << "  static constexpr std::uint32_t " << since_version_constant_name() << " = " << since << ";" << std::endl;
 
@@ -268,7 +277,7 @@ struct request_t : public event_t
       ss <<  "void ";
     else
       ss << ret.print_type() << " ";
-    ss << interface_name << "_t::" << name << "(";
+    ss << interface_name << "_t::" << sanitise(name) << "(";
 
     bool new_id_arg = false;
     for(auto &arg : args)
@@ -308,13 +317,13 @@ struct request_t : public event_t
             ss << "nullptr, ";
           }
         else if(arg.type == "fd")
-          ss << "argument_t::fd(" << arg.name << "), ";
+          ss << "argument_t::fd(" << sanitise(arg.name) << "), ";
         else if(arg.type == "object")
-          ss << arg.name << ".proxy_has_object() ? reinterpret_cast<wl_object*>(" << arg.name << ".c_ptr()) : nullptr, ";
+          ss << sanitise(arg.name) << ".proxy_has_object() ? reinterpret_cast<wl_object*>(" << sanitise(arg.name) << ".c_ptr()) : nullptr, ";
         else if(arg.enum_name != "")
-          ss << "static_cast<" << arg.print_enum_wire_type() << ">(" << arg.name + "), ";
+          ss << "static_cast<" << arg.print_enum_wire_type() << ">(" << sanitise(arg.name) + "), ";
         else
-          ss << arg.name + ", ";
+          ss << sanitise(arg.name) + ", ";
       }
 
     ss.str(ss.str().substr(0, ss.str().size()-2));
@@ -348,13 +357,11 @@ struct request_t : public event_t
 
 struct enum_entry_t : public element_t
 {
-  std::string name;
   std::string value;
 };
 
 struct enumeration_t : public element_t
 {
-  std::string name;
   std::list<enum_entry_t> entries;
   bool bitfield;
   int id;
@@ -394,9 +401,9 @@ struct enumeration_t : public element_t
           ss << "  /** \\brief " << entry.summary << " */" << std::endl;
 
         if(!bitfield)
-          ss << "  " << entry.name << " = " << entry.value << "," << std::endl;
+          ss << "  " << sanitise(entry.name) << " = " << entry.value << "," << std::endl;
         else
-          ss << "  static const detail::bitfield<" << width << ", " << id << "> " << entry.name << ";" << std::endl;
+          ss << "  static const detail::bitfield<" << width << ", " << id << "> " << sanitise(entry.name) << ";" << std::endl;
       }
 
     if(!bitfield)
@@ -417,7 +424,7 @@ struct enumeration_t : public element_t
       for(auto &entry : entries)
         {
           ss << "const bitfield<" << width << ", " << id << "> " << iface_name << "_" << name
-             << "::" << entry.name << "{" << entry.value << "};" << std::endl;
+             << "::" << sanitise(entry.name) << "{" << entry.value << "};" << std::endl;
         }
     return ss.str();
   }
@@ -426,7 +433,6 @@ struct enumeration_t : public element_t
 struct interface_t : public element_t
 {
   int version;
-  std::string name;
   std::string orig_name;
   int destroy_opcode;
   std::list<request_t> requests;
@@ -690,22 +696,52 @@ std::string unprefix(const std::string &name)
   return name;
 }
 
+struct arg_t
+{
+  std::string key;
+  std::string value;
+};
+
+void parse_args(int argc, char **argv, std::vector<arg_t>& map, std::vector<std::string>& extra)
+{
+  bool opts_end = false;
+  for(unsigned int c = 1; c < argc; c++)
+  {
+    std::string str(argv[c]);
+    if(opts_end || str[0] != '-')
+      extra.push_back(str);
+    else if(str == "--")
+      opts_end = true;
+    else
+    {
+      std::string value;
+      if (c + 1 < argc && argv[c+1][0] != '-')
+	value = argv[++c];
+      map.push_back(arg_t{str.substr(1), value});
+    }
+  }
+}
+
 int main(int argc, char *argv[])
 {
-  if(argc < 4)
+  std::vector<arg_t> map;
+  std::vector<std::string> extra;
+  parse_args(argc, argv, map, extra);
+
+  if(extra.size() < 3)
     {
       std::cerr << "Usage:" << std::endl
-                << "  " << argv[0] << " protocol1.xml [protocol2.xml ...] protocol.hpp protocol.cpp" << std::endl;
+                << "  " << argv[0] << " [-x extra_header.hpp] protocol1.xml [protocol2.xml ...] protocol.hpp protocol.cpp" << std::endl;
       return 1;
     }
 
   std::list<interface_t> interfaces;
   int enum_id = 0;
 
-  for(int c = 1; c < argc-2; c++)
+  for(int c = 0; c < extra.size()-2; c++)
     {
       xml_document doc;
-      doc.load_file(argv[c]);
+      doc.load_file(extra[c].c_str());
       xml_node protocol = doc.child("protocol");
 
       for(xml_node &interface : protocol.children("interface"))
@@ -889,7 +925,7 @@ int main(int argc, char *argv[])
         }
     }
 
-  std::string hpp_file(argv[argc-2]), cpp_file(argv[argc-1]);
+  std::string hpp_file(extra[extra.size()-2]), cpp_file(extra[extra.size()-1]);
   std::fstream wayland_hpp(hpp_file, std::ios_base::out | std::ios_base::trunc);
   std::fstream wayland_cpp(cpp_file, std::ios_base::out | std::ios_base::trunc);
 
@@ -902,8 +938,13 @@ int main(int argc, char *argv[])
               << "#include <string>" << std::endl
               << "#include <vector>" << std::endl
               << std::endl
-              << "#include <wayland-client.hpp>" << std::endl
-              << std::endl;
+              << "#include <wayland-client.hpp>" << std::endl;
+
+  for(auto const& opt : map)
+    if(opt.key == std::string("x"))
+      wayland_hpp << "#include <" << opt.value << ">" << std::endl;
+
+  wayland_hpp << std::endl;
 
   // C forward declarations
   for(auto &iface : interfaces)
@@ -960,3 +1001,92 @@ int main(int argc, char *argv[])
   
   return 0;
 }
+
+// set of C++-only keywords not to use as names
+const std::set<std::string> element_t::keywords =
+  {
+   "alignas",
+   "alignof",
+   "and",
+   "and_eq",
+   "asm",
+   "auto",
+   "bitand",
+   "bitor",
+   "bool",
+   "break",
+   "case",
+   "catch",
+   "char",
+   "char16_t",
+   "char32_t",
+   "class",
+   "compl",
+   "const",
+   "constexpr",
+   "const_cast",
+   "continue",
+   "decltype",
+   "default",
+   "delete",
+   "do",
+   "double",
+   "dynamic_cast",
+   "else",
+   "enum",
+   "explicit",
+   "export",
+   "extern",
+   "false",
+   "float",
+   "for",
+   "friend",
+   "goto",
+   "if",
+   "inline",
+   "int",
+   "long",
+   "mutable",
+   "namespace",
+   "new",
+   "noexcept",
+   "not",
+   "not_eq",
+   "nullptr",
+   "operator",
+   "or",
+   "or_eq",
+   "private",
+   "protected",
+   "public",
+   "register",
+   "reinterpret_cast",
+   "return",
+   "short",
+   "signed",
+   "sizeof",
+   "static",
+   "static_assert",
+   "static_cast",
+   "struct",
+   "switch",
+   "template",
+   "this",
+   "thread_local",
+   "throw",
+   "true",
+   "try",
+   "typedef",
+   "typeid",
+   "typename",
+   "union",
+   "unsigned",
+   "using",
+   "virtual",
+   "void",
+   "volatile",
+   "wchar_t",
+   "while",
+   "xor",
+   "xor_eq",
+  };
