@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2014-2017, Nils Christopher Brause
+ * Copyright (c) 2014-2019, Nils Christopher Brause, Philipp Kerling
  * All rights reserved.
  * 
  * Redistribution and use in source and binary forms, with or without
@@ -453,6 +453,14 @@ display_t::display_t(std::string name)
   interface = &display_interface;
 }
 
+display_t::display_t(wl_display* display)
+  : proxy_t(reinterpret_cast<wl_proxy*> (display), proxy_t::wrapper_type::foreign)
+{
+  if(!proxy_has_object())
+    throw std::runtime_error("Cannot construct display_t wrapper from nullptr");
+  interface = &display_interface;
+}
+
 display_t::display_t(display_t &&d)
 {
   operator=(std::move(d));
